package xyz.matj1.bp;

import mcdr.sequence.SequenceMocap;
import mcdr.sequence.SequenceMotionWords;
import messif.objects.LocalAbstractObject;
import smf.modules.distance.DTWSequenceDist;
import smf.modules.distance.DTWSequenceDistNormalized;
import smf.sequences.Sequence;
import xyz.matj1.bp.enums.MocapType;

import java.util.function.BiFunction;

public class DistanceFunctions {
	public static class Basic implements BiFunction<LocalAbstractObject, LocalAbstractObject, Double> {
		@Override
		public Double apply(LocalAbstractObject o1, LocalAbstractObject o2) {
			return basic(o1, o2);
		}
	}

	public static class CompensedByDiagonal implements BiFunction<LocalAbstractObject, LocalAbstractObject, Double> {
		@Override
		public Double apply(LocalAbstractObject o1, LocalAbstractObject o2) {
			return compensedByDiagonal(o1, o2);
		}
	}

	public static class CompensedByDiagonalFromUri implements BiFunction<LocalAbstractObject, LocalAbstractObject, Double> {
		@Override
		public Double apply(LocalAbstractObject o1, LocalAbstractObject o2) {
			return compensedByDiagonalFromUri(o1, o2);
		}
	}

	public static class CompensedByWarpingPath implements BiFunction<LocalAbstractObject, LocalAbstractObject, Double> {
		@Override
		public Double apply(LocalAbstractObject o1, LocalAbstractObject o2) {
			return compensedByWarpingPath(o1, o2);
		}
	}

	public static Double basic(LocalAbstractObject a, LocalAbstractObject b) {
		return (double)a.getDistance(b);
	}

	public static <T> Double compensedByDiagonal(LocalAbstractObject a, LocalAbstractObject b) {
		try {
			var aSeq = (Sequence<Object>)a;
			var bSeq = (Sequence<Object>)b;

			var distFunction = new DTWSequenceDist<Object>();
			var distance = distFunction.getDistance(aSeq, bSeq);

			var lengthA = aSeq.getSequenceLength();
			var lengthB = bSeq.getSequenceLength();

			return distance / Math.sqrt(lengthA * lengthA + lengthB * lengthB);
		} catch (ClassCastException e) {
			throw new IllegalArgumentException("A parameter can't be cast to Sequence", e);
		}
	}
	
	/**
	 * This computes the distance of two sequences compensed by their lengths.
	 * The sequences should be of the same {@link MocapType},
	 * otherwise the result probably wouldn't make sense.
	 * Both sequences have to be instances of {@link SequenceMocap} or
	 * both sequences have to be instances of {@link SequenceMotionWords}.
	 *
	 * The compensation is that the raw distance is divided by the length
	 * of the diagonal of a rectangle with the sequences as the sides.
	 * The lengths of their sequences are taken from their URIs.
	 *
	 * @param a the first sequence
	 * @param b the second sequence
	 * @return their distance compensation for their lengths.
	 */
	public static Double compensedByDiagonalFromUri(LocalAbstractObject a, LocalAbstractObject b) {
		var distance = a.getDistance(b);

		var lengthA = Utils.getLengthFromUri(a);
		var lengthB = Utils.getLengthFromUri(b);

		return distance / Math.sqrt(lengthA*lengthA + lengthB*lengthB);
	}

	public static Double compensedByWarpingPath(LocalAbstractObject a, LocalAbstractObject b) {
		try {
			var distFunction = new DTWSequenceDistNormalized<Object>();
			return (double)distFunction.getDistance((Sequence)a, (Sequence)b);
		} catch (ClassCastException e) {
			throw new IllegalArgumentException("A parameter can't be cast to Sequence", e);
		}
	}
}
