package xyz.matj1.bp;

import mcdr.sequence.SequenceMotionWords;
import mcdr.test.utils.ObjectCategoryMgmt.Category;
import mcdr.test.utils.ObjectMgmt;
import messif.objects.LocalAbstractObject;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.data.category.DefaultCategoryDataset;
import org.jfree.data.statistics.DefaultBoxAndWhiskerCategoryDataset;
import org.jfree.data.xy.DefaultXYDataset;
import xyz.matj1.bp.enums.MocapType;
import xyz.matj1.bp.enums.SpeedMultiple;
import xyz.matj1.bp.exceptions.CacheIOException;
import xyz.matj1.bp.experiments.KnnPrecision;
import xyz.matj1.bp.types.Category_Multiple_Precision;
import xyz.matj1.bp.types.MyMap;
import xyz.matj1.bp.types.PointSet;
import xyz.matj1.bp.types.Vector2;

import java.io.File;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.Writer;
import java.time.Duration;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.function.Predicate;
import java.util.regex.PatternSyntaxException;
import java.util.stream.DoubleStream;
import java.util.stream.IntStream;

import static xyz.matj1.bp.Utils.average;
import static xyz.matj1.bp.Utils.loadData;
import static xyz.matj1.bp.Utils.readCache;
import static xyz.matj1.bp.Utils.writeCache;
import static xyz.matj1.bp.Utils.writeXYGraphAsPdf;
import static xyz.matj1.bp.enums.SpeedMultiple._1;

/**
 * A class for testing of how motion-word sequences
 * can be matched on artificially stretched sequences
 *
 * @author Matěj Bagar
 */
public class StretchedSequences {
	/**
	 * This runs tests on stretched sequences.
	 *
	 * @param args
	 * @throws IOException
	 */
	public static void main(String[] args) throws IOException {
//		knnTest(new OutputStreamWriter(System.out));
//		knnTestShow(new File(Config.outputDir(), "stretch kNN averages"));
		knnTestUnified(new OutputStreamWriter(System.out));
//		knnTestUnifiedShow(Path.of(Config.outputDir().getPath(), "stretch kNN averages", "unified graph.pdf").toFile());
//		knnTestUnifiedShowBox(Path.of(Config.outputDir().getPath(), "stretch kNN averages", "unified box graph.pdf").toFile());
//		stretchingCheckWrite(new FileWriter(new File(Config.outputDir(), "stretchingCheckWrite")));
//		stretchingCheckAverageWrite(new OutputStreamWriter(System.out));
//		stretchingCheckAverageShow(new File(Config.outputDir(), "stretch check averages"));
//		distanceFunctionTimingTest(new PrintWriter(System.out));
	}

	/**
	 * This computes the average precision for each category, multiple, type.
	 *
	 * @param output where the output shall be written
	 * @throws IOException
	 */
	private static void knnTest(Writer output) throws IOException {
		var knnMeasurements = getKnnMeasurements();

		for (var type : MocapType.values()) {
			output.write(String.format("%s:%n", type));
			for (var multiple : SpeedMultiple.values()) {
				output.write(String.format("%s:%n", multiple.asString()));
				Utils.write(
					KnnPrecision.averageFormat(knnPrecisionAverage(
						knnMeasurements.get(type).get(multiple)
					)),
					output
				);
			}
			output.write(System.lineSeparator());
	}}

	/**
	 * This computes the average precision for each category, multiple, type, and plots it to graphs.
	 *
	 * @param outputDir where the output files shall be put
	 * @throws IOException if the reading or writing fails
	 */
	private static void knnTestShow(File outputDir) throws IOException {
		outputDir.mkdirs();
		var knnMeasurements = getKnnMeasurements();

		for (var type : MocapType.values()) {
			var precisions = new Category_Multiple_Precision();
			for (var multiple : SpeedMultiple.values()) {
				var category_List = knnMeasurements.get(type).get(multiple);

				for (var category : category_List.keySet()) {
					precisions.addPrecision(category, multiple, average(category_List.get(category)));
				}
			}

			var dataset = new DefaultCategoryDataset();
			var sortedCategoryEntries = precisions.entrySet().stream()
				.sorted(Comparator.comparing(e -> e.getValue().get(_1)))
				.toList();

			for (var entry : sortedCategoryEntries) {
				for (var multiple : SpeedMultiple.values()) {
					dataset.addValue(entry.getValue().get(multiple), multiple, entry.getKey().id);
				}
			}

			var graph = ChartFactory.createLineChart(
				String.format("Average kNN precision of each sequence in each multiple among the original sequences as %s", type),
				"category",
				"average precision",
				dataset,
				PlotOrientation.VERTICAL,
				true, true, true
			);
			writeXYGraphAsPdf(
				graph,
				new File(outputDir, String.format("%s.pdf", type)),
				new Vector2<>(1024, 512)
			);
		}
	}

	/**
	 * This shows a bar graph with the average kNN precision for each speed multiple for each mocap type.
	 *
	 * @param output where the output shall be written
	 * @throws IOException if the reading or writing fails
	 */
	private static void knnTestUnified(Writer output) throws IOException {
		var knnMeasurements = getKnnMeasurements();

		for (var type : MocapType.values()) {
			for (var multiple : SpeedMultiple.values()) {
				var category_List = knnMeasurements.get(type).get(multiple);
				DoubleStream flatValues = category_List.values().stream()
					.flatMapToDouble(
						x -> x.stream().mapToDouble(a -> a).filter(Double::isFinite)
					);
				var average = flatValues.average().orElseThrow();

				output.write(String.format(
					"%s, %s×: %s%n",
					type, multiple, average
				));
				output.flush();
			}
		}
	}


	/**
	 * This shows a bar graph with the average kNN precision for each speed multiple for each mocap type.
	 *
	 * @param outputFile where the output shall be written
	 * @throws IOException if the reading or writing fails
	 */
	private static void knnTestUnifiedShow(File outputFile) throws IOException {
		outputFile.getParentFile().mkdirs();
		var knnMeasurements = getKnnMeasurements();
		var dataset = new DefaultCategoryDataset();

		for (var type : MocapType.values()) {
			for (var multiple : SpeedMultiple.values()) {
				var category_List = knnMeasurements.get(type).get(multiple);
				DoubleStream flatValues = category_List.values().stream()
					.flatMapToDouble(
						x -> x.stream().mapToDouble(a -> a)
					);
				var average = flatValues.average().orElseThrow();

				dataset.addValue(average, multiple, type);
			}
		}
		var graph = ChartFactory.createBarChart(
			"Average kNN precision for each speed multiple among the original sequences for each mocap type",
			"type",
			"average kNN precision",
			dataset,
			PlotOrientation.VERTICAL,
			true, true, true
		);
		writeXYGraphAsPdf(
			graph,
			outputFile,
			new Vector2<>(1024, 512)
		);
	}

	/**
	 * This shows a box graph for each speed multiple for each mocap type.
	 *
	 * @param outputFile where the output shall be written
	 * @throws IOException if the reading or writing fails
	 */
	private static void knnTestUnifiedShowBox(File outputFile) throws IOException {
		outputFile.getParentFile().mkdirs();
		var knnMeasurements = getKnnMeasurements();
		var dataset = new DefaultBoxAndWhiskerCategoryDataset();

		for (var type : MocapType.values()) {
			for (var multiple : SpeedMultiple.values()) {
				var category_List = knnMeasurements.get(type).get(multiple);
				var flatValues = category_List.values().stream()
					.flatMap(Collection::stream)
					.toList();

				dataset.add(flatValues, multiple, type);
			}
		}
		var graph = ChartFactory.createBoxAndWhiskerChart(
			"Average kNN precision for each speed multiple among the original sequences for each mocap type",
			"type",
			"average kNN precision",
			dataset,
			true
		);
		writeXYGraphAsPdf(
			graph,
			outputFile,
			new Vector2<>(1024, 512)
		);

	}

	/**
	 * This computes the distance from each sequence to its original sequence.
	 * The distances are indexed by mocap type, speed multiple, then by category.
	 *
	 * @return
	 * @throws IOException
	 */
	private static MyMap<MocapType, MyMap<SpeedMultiple, MyMap<Category, List<Double>>>> stretchingCheck() throws IOException {
		var cacheFile = new File(Config.cacheDir(), "stretchingCheck");
		try {
			return readCache(cacheFile, MyMap.class);
		} catch (CacheIOException _e) {
			var result = new MyMap<MocapType, MyMap<SpeedMultiple, MyMap<Category, List<Double>>>>();

			for (var type : MocapType.values()) {
				result.put(type, new MyMap<>());
				var typeMap = result.get(type);

				var category_OriginalSequences = loadData(dataFiles(type).get(_1), type.dataClass()).getCategoryObjects();
				for (var multiple : SpeedMultiple.values()) {
					typeMap.put(multiple, new MyMap<>());
					var multipleMap = typeMap.get(multiple);

					var stretchedSequenceMgmt = loadData(dataFiles(type).get(multiple), type.dataClass());
					var category_StretchedSequences = stretchedSequenceMgmt.getCategoryObjects();

					stretchedSequenceMgmt.getCategories().parallelStream().forEach(category -> {
						multipleMap.put(category, new ArrayList<>());
						var categoryList = multipleMap.get(category);

						category_StretchedSequences.get(category).parallelStream()
							.forEach(sequence ->  {
								var originalSequence = category_OriginalSequences.get(category)
									.stream().filter(uriMatches(sequence)).findAny().orElseThrow();
								var distance = DistanceFunctions.compensedByDiagonalFromUri(sequence, originalSequence);
								categoryList.add(distance);
							});
					});
			}}

			try {
				writeCache(cacheFile, result);
			} catch (CacheIOException e) {
				System.err.printf("Writing cache to %s failed.%n", cacheFile);
				e.printStackTrace();
			}

			return result;
		}
	}

	/**
	 * This is to check that every stretched sequence
	 * has a very small distance to its original sequence.
	 * This writes the length-compensed distance from each sequence to its original sequence
	 * to <code>output</code>.
	 *
	 * @param output where to write the results
	 * @throws IOException if loading the data fails
	 */
	private static void stretchingCheckWrite(Writer output) throws IOException {
		var result = stretchingCheck();

		for (var type : result.keySet()) {
			output.write(String.format("%s:%n", type));
			var multiple_Map = result.get(type);
			for (var multiple : multiple_Map.keySet()) {
				output.write(String.format("%s:%n", multiple));
				var category_Map = multiple_Map.get(multiple);
				for (var category : category_Map.keySet()) {
					output.write(String.format("Category: %s:%n", category.id));
					var distanceList = category_Map.get(category);
					for (var distance : distanceList) {
						output.write(String.format("%s%n", distance));
		}}}}
		output.flush();
		System.err.printf("Stretch checking results were written to %s.%n", output);
	}

	/**
	 * This is to check that every stretched sequence
	 * has a very small distance to its original sequence.
	 * This writes the length-compensed distance from each sequence to its original sequence
	 * to <code>output</code>.
	 *
	 * @param output where to write the results
	 * @throws IOException if loading the data fails
	 */
	private static void stretchingCheckAverageWrite(Writer output) throws IOException {
		var result = stretchingCheck();

		for (var type : result.keySet()) {
			output.write(String.format("%s:%n", type));
			var multiple_Map = result.get(type);
			for (var multiple : multiple_Map.keySet()) {
				output.write(String.format("%s:%n", multiple));
				var category_Map = multiple_Map.get(multiple);
				for (var category : category_Map.keySet()) {
					var averageDistance = average(category_Map.get(category));
					output.write(String.format("Category %s: %s%n", category.id, averageDistance));
		}}}
		output.flush();
		System.err.printf("Stretch checking averages were written to %s.%n", output);
	}

	/**
	 * This is to check that every stretched sequence
	 * has a very small distance to its original sequence.
	 *
	 * @param outputDir where to write the results
	 * @throws IOException if loading the data or writing the results fails
	 */
	private static void stretchingCheckAverageShow(File outputDir) throws IOException {
		outputDir.mkdirs();
		var result = stretchingCheck();

		for (var type : result.keySet()) {
			var multiple_Map = result.get(type);
			for (var multiple : multiple_Map.keySet()) {
				var category_List = multiple_Map.get(multiple);
				var pointSet = new PointSet();

				for (var category : category_List.keySet()) {
					var averageDistance = average(category_List.get(category));
					pointSet.add(Double.parseDouble(category.id), averageDistance);
				}

				var dataset = new DefaultXYDataset();
				dataset.addSeries(0, pointSet.toDataSeries());
				var graph = ChartFactory.createScatterPlot(
					String.format("Average distance of each sequence stretched %s× to its original sequence as %s", multiple, type),
					"category",
					"average distance",
					 dataset,
					PlotOrientation.VERTICAL,
					false, true, true
				);
				writeXYGraphAsPdf(
					graph,
					new File(outputDir, String.format("%s – %s×.pdf", type, multiple)),
					new Vector2<>(1024, 512)
				);
	}}}

	private static void distanceFunctionTimingTest(PrintWriter output) throws IOException {
		var mocapType = MocapType.HARD_MW;
		var mocapClass = SequenceMotionWords.class;

		var sequenceMgmt = loadData(dataFiles(mocapType).get(SpeedMultiple._1), mocapType.dataClass());
		var sequences = sequenceMgmt.getObjects().stream().toList();
		var firstSequence = sequences.get(0);

		var basicSeqStream = sequences.stream();
		var basicTimer = Instant.now();
		var basicList = basicSeqStream.map(
			o -> o.getDistance(firstSequence)
		).toList();
		var basicTime = Duration.between(basicTimer, Instant.now());
		System.err.println(basicList.size()); // to prevent the distance computations from being optimised out
		output.printf("Time for basic distances: %s ms%n", basicTime.toMillis());

		var diagonalSeqStream = sequences.stream();
		var diagonalTimer = Instant.now();
		var diagonalList = diagonalSeqStream.map(
			o -> DistanceFunctions.compensedByDiagonal(o, firstSequence)
		).toList();
		var diagonalTime = Duration.between(diagonalTimer, Instant.now());
		System.err.println(diagonalList.size()); // to prevent the distance computations from being optimised out
		output.printf("Time for distances compensed by diagonal: %s ms%n", diagonalTime.toMillis());

		var diagonalUriSeqStream = sequences.stream();
		var diagonalUriTimer = Instant.now();
		var diagonalUriList = diagonalUriSeqStream.map(
			o -> DistanceFunctions.compensedByDiagonalFromUri(o, firstSequence)
		).toList();
		var diagonalUriTime = Duration.between(diagonalUriTimer, Instant.now());
		System.err.println(diagonalUriList.size()); // to prevent the distance computations from being optimised out
		output.printf("Time for distances compensed by diagonal from URI: %s ms%n", diagonalUriTime.toMillis());

		var warpingPathSeqStream = sequences.stream();
		var warpingPathTimer = Instant.now();
		var warpingPathList = warpingPathSeqStream.map(
			o -> DistanceFunctions.compensedByWarpingPath(o, firstSequence)
		).toList();
		var warpingPathTime = Duration.between(warpingPathTimer, Instant.now());
		System.err.println(warpingPathList.size()); // to prevent the distance computations from being optimised out
		output.printf("Time for distances compensed by warping path: %s ms%n", warpingPathTime.toMillis());

		output.flush();
	}

	/**
	 * This returns a {@link Predicate} that tells if its given {@link LocalAbstractObject}
	 * has a locator URI matching to the locator URI of <code>sequence</code>.
	 * Two LocalAbstractObjects have matching locator URIs iff
	 * the URIs are digit sequences separated by “_” and the a three of them are equal.
	 * This is supposed to be used to search for matching stretched sequences.
	 *
	 * @param sequence a sequence to whose URI the other's URI shall be matched
	 * @return a predicate that tells if its given object has a locator URI matching to that of <code>sequence</code>
	 */
	public static Predicate<LocalAbstractObject> uriMatches(LocalAbstractObject sequence) {
		return other -> {
			try {
				var uriArray = sequence.getLocatorURI().split("_");
				var otherUriArray = other.getLocatorURI().split("_");

				return IntStream.of(0, 1, 2).allMatch(
					x -> uriArray[x].equals(otherUriArray[x])
				);
			} catch (PatternSyntaxException | NumberFormatException | IndexOutOfBoundsException e) {
				return false;
		}};
	}

	/**
	 * This takes the output from {@link KnnPrecision#compute(ObjectMgmt, ObjectMgmt)}
	 * and creates an average of precision for each category.
	 *
	 * It's very similar to {@link KnnPrecision#average(Map)};
	 * the only difference is that that function takes data containing the sequences and ignores them,
	 * and this function takes data without the sequences.
	 *
	 * @param knnData an output from {@link KnnPrecision#compute(ObjectMgmt, ObjectMgmt)}
	 * @return a map from each category to the average precision in the category
	 */
	public static MyMap<Category, Double> knnPrecisionAverage(MyMap<Category, List<Double>> knnData) {
		return knnData
			.mapValues(
				value -> value.parallelStream().mapToDouble(x -> x).average().orElse(Double.NaN)
			);
	}

	/**
	 * This gets, for every mocap type, for every speed multiple, for every category,
	 * for every sequence in the category, the precision of kNN search
	 * of the sequence with the speed multiple among sequences with multiple 1.
	 *
	 * This either computes it or loads it from cache.
	 *
	 * @return a map of every mocap type to a map of every speed multiple to a map of every category to a list of all kNN precisions of its sequences
	 * @throws IOException
	 */
	public static MyMap<MocapType, MyMap<SpeedMultiple, MyMap<Category, List<Double>>>> getKnnMeasurements() throws IOException {
		var cacheDir = new File(Config.cacheDir(), "stretchKnnMeasurements");
		cacheDir.mkdirs();

		var knnMeasurements = new MyMap<MocapType, MyMap<SpeedMultiple, MyMap<Category, List<Double>>>>();

		for(var type : MocapType.values()) {
			var multiple_Measurements = new MyMap<SpeedMultiple, MyMap<Category, List<Double>>>();
			for(var multiple : SpeedMultiple.values()) {
				var cachePath = new File(cacheDir, String.format("%s – %s×", type.noun(), multiple.asString()));

				try {
					multiple_Measurements.put(multiple, readCache(cachePath, MyMap.class));
				} catch(CacheIOException _e) {
					var stretchedSequenceMgmt = loadData(dataFiles(type).get(multiple), type.dataClass());
					var baseSequenceMgmt = loadData(dataFiles(type).get(_1), type.dataClass());
					multiple_Measurements.put(multiple, KnnPrecision.compute(baseSequenceMgmt, stretchedSequenceMgmt));

					try {
						writeCache(cachePath, multiple_Measurements.get(multiple));
					} catch (CacheIOException e) {
						System.err.printf("Writing cache to %s failed.%n", cachePath);
						e.printStackTrace();
					}
				}
			}

			var correctedMultiple_Measurements = multiple_Measurements.mapValues(
				inmap -> inmap.mapValues(
					list -> list.stream().map(
						value -> value.isInfinite() ? Double.NaN : value
					).toList()
			)); // to correct all infinite values from the divisions by zero during the computations for the categories with 1 as the size
			// TODO: remove this after rebuilding cache; the problem in KnnPrecision#compute(ObjectMgmt, ObjectMgmt) is fixed

			knnMeasurements.put(type, correctedMultiple_Measurements);
		}

		return knnMeasurements;
	}

	private static MyMap<SpeedMultiple, File> dataFiles(MocapType type) {
		var speedMultiples = Arrays.asList(SpeedMultiple.values());

		return switch (type) {
			case SKELETON -> speedMultiples.stream().collect(MyMap.keyCollector(
				mult -> new File(
					Config.dataDir(),
					"skeleta/class130-actions-coords_normPOS" + (mult.equals(_1) ? "" : "_stretched-" + mult.asString() + "×")
						+ ".data"
			)));

			case HARD_MW -> speedMultiples.stream().collect(MyMap.keyCollector(
				mult -> new File(
					Config.dataDir(),
					"motion words/hard_motion_words"
						+ (mult.equals(_1) ? "" : "_stretched_" + mult.asString() + "×")
						+ ".data"
			)));

			case SOFT_MW -> speedMultiples.stream().collect(MyMap.keyCollector(
				mult -> new File(
					Config.dataDir(),
					"motion words/soft_motion_words"
						+ (mult.equals(_1) ? "" : "_stretched_" + mult.asString() + "×")
						+ ".data"
			)));

			case MATCH_N_MW -> speedMultiples.stream().collect(MyMap.keyCollector(
				mult -> new File(
					Config.dataDir(),
					"motion words/multi-overlay_motion_words"
						+ (mult.equals(_1) ? "" : "_stretched_" + mult.asString() + "×")
						+ ".data"
			)));
		};
	}
}
